//
// o p t u s b . c
//
// Multiboot program downloader for optimize's USB-bootcable
//
// Version 1.01+darwin
//
// Target OS: Mac OS X 10.3
// Build: cc -O2 -I/usr/local/include -c -o optusb.o optusb.c
//        cc -o optusb optusb.o -L/usr/local/lib -lusb -lIOKit
//
// NOTICE: This version also requires the "libusb" library (CVS version)
//
// Nov 22, 2003 by gbcube (http://home.r00.itscom.net/top/gbcube/)
//
// Version 1.0+darwin
//
// Target OS: Mac OS X 10.2.6
// Build: cc -O2 -I./libusb/include -c -o optusb.o optusb.c
//        cc -o optusb optusb.o -L./libusb/lib -lusb -lIOKit
//
// NOTICE: This version also requires the "libusb" library (CVS version)
//         http://libusb.sourceforge.net/
//
// June 22, 2003 by gbcube (http://www.geocities.co.jp/SiliconValley/9274/)
//
// Version 1.0
//
// Target OS: Linux-x86-pc
// Build: gcc -Wall -o optusb optusb.c libusb.a
//
// NOTICE: This program requires the "libusb" library
//         http://libusb.sourceforge.net/
//
// June 16, 2003 by W.Nishida (http://www.skyfree.org)
//

#include <stdio.h>	// fprintf(), stderr
#include <string.h>	// strcmp()
#include <unistd.h>	// open(), read(), close(), sleep()
#include <fcntl.h>	// O_RDONLY
#include <time.h>	// nanosleep(), struct timespec
#include "usb.h"	// "libusb" header file

// Device descriptor for USB-bootcable

#define OPTUSB_VEND	0x0BFE		// Vendor code
#define OPTUSB_PROD	0x3000		// Product code

// Endpoints in USB-bootcable

#define OPTUSB_CMD	0x01		// Command endpoint
#define OPTUSB_IN	0x82		// Input endpoint
#define OPTUSB_OUT	0x02		// Output endpoint

// Commands required for the program transfer

#define OPTUSB_WRITE	2		// Write data
#define OPTUSB_STATUS	3		// Read status

// Negotiation codes

#define OPTUSB_MAGIC	0xFEDCBA98	// Magic number for negotiation test
#define OPTUSB_ON	0x01		// Status code bit0: GBA power
#define OPTUSB_READY	0x02		// Status code bit1: ready to load

// Constant

#define MAX_PROG_SIZE	(1024 * 256)	// 256KB

// Global variable

struct usb_dev_handle *optusb;		// Device handler for bootcable

// msleep

static void Sleep(int ms)
{
  struct timespec ts;
  ts.tv_sec = ms / 1000;
  ts.tv_nsec = (ms % 1000) * 1000000;
  nanosleep(&ts, NULL);
}

// Read a magic number

int read_magic(int num) {
  char cmd[ 2 ];
  unsigned char val[4];
  int ret;

  cmd[ 0 ] = OPTUSB_STATUS;
  cmd[ 1 ] = num;
  ret = usb_bulk_write(optusb, OPTUSB_CMD, cmd, sizeof(cmd), 1000);
  if (ret != sizeof(cmd))
    return -1;
  ret = usb_bulk_read(optusb, OPTUSB_IN, (char*)val, sizeof(val), 1000);
  if (ret != sizeof(val))
    return -1;
  return val[3]<<24 | val[2]<<16 | val[1]<<8 | val[0];
 }

// Send a magic number

int send_magic(int num, int magic) {
  char cmd[ 6 ];
  unsigned char val[4];
  int ret;

  cmd[ 0 ] = OPTUSB_STATUS;
  cmd[ 1 ] = 0x80 + num;
  cmd[ 2 ] = magic & 0xFF;
  cmd[ 3 ] = (magic >>  8) & 0xFF;
  cmd[ 4 ] = (magic >> 16) & 0xFF;
  cmd[ 5 ] = (magic >> 24) & 0xFF;

  ret = usb_bulk_write(optusb, OPTUSB_CMD, cmd, sizeof(cmd), 1000);
  if (ret != sizeof(cmd))
    return -1;
  ret = usb_bulk_read(optusb, OPTUSB_IN, (char*)val, sizeof(val), 1000);
  if (ret != sizeof(val))
    return -1;
  return val[3]<<24 | val[2]<<16 | val[1]<<8 | val[0];
 }

// Send buffer data via USB

int send_file(char *buf, int len) {
  char cmd[ 4 ];
  int n;
  int ret, sizew;

  // Send program length

  n = 0;
  cmd[ n++ ] = OPTUSB_WRITE;
  cmd[ n++ ] = 1;
  cmd[ n++ ] = 0;
  ret = usb_bulk_write(optusb, OPTUSB_CMD, cmd, n, 1000);
  if (ret != n)
    return -1;
  Sleep(50);

  n = 0;
  cmd[ n++ ] = len & 0xFF;
  cmd[ n++ ] = (len >> 8) & 0xFF;
  cmd[ n++ ] = (len >> 16) & 0xFF;
  cmd[ n++ ] = (len >> 24) & 0xFF;
  ret = usb_bulk_write(optusb, OPTUSB_OUT, cmd, n, 1000);
  if (ret != n)
    return -1;

  // Send data

  sizew = ((len + 3) / 4) * 4;		// Program size in word boundary
  n = 0;
  cmd[ n++ ] = OPTUSB_WRITE;
  cmd[ n++ ] = (sizew >> 2) & 0xFF;
  cmd[ n++ ] = (sizew >> 10) & 0xFF;
  ret = usb_bulk_write(optusb, OPTUSB_CMD, cmd, n, 1000);
  if (ret != n)
    return -1;

  ret = usb_bulk_write(optusb, OPTUSB_OUT, buf, sizew, 1000);
  if (ret != sizew)
    return -1;
    
  Sleep(100);

  return 0;
 }

int main(int argc, char** argv) {
  struct usb_bus *bus, *busses;
  struct usb_device *dev;
  struct usb_device_descriptor *dd;
  unsigned char *buf;
  int ret, fd, len;
  char *mbfname;

  // Check command argument

  if (argc <= 1) {
    fprintf(stderr, "Usage: optusb [-v] mb-file-name\n");
    return 1;
   }
  if (strcmp(argv[ 1 ], "-v") == 0) {
    if (argc == 3) {
      usb_set_debug(3);
      mbfname = argv[ 2 ];
     }
    else {
      fprintf(stderr, "  Error: mb-file-name not specified!\n");
      return 1;
     }
   }
  else {
    mbfname = argv[ 1 ];
   }
  fprintf(stderr, "      === optusb v1.01+darwin ===\n\n");

  // Allocate 256KB buffer

  buf = malloc(MAX_PROG_SIZE);
  if (buf == NULL) {
    fprintf(stderr, "  Error: Insufficient memory!\n");
    return 1;
   }

  // Open a program file

  fd = open(mbfname, O_RDONLY);
  if (fd == -1) {
    fprintf(stderr, "  Error: file open failure!\n");
    return 1;
   }

  // Read into all data

  len = read(fd, buf, MAX_PROG_SIZE);
  if (len == 0) {
    fprintf(stderr, "  Error: file read error!\n");
    return 1;
   }

  fprintf(stdout, "   Source file = %s\n", mbfname);
  fprintf(stdout, "  Program size = %d\n", len);

  // Initialize "libusb" and get device information

  usb_init();

  usb_find_busses();
  usb_find_devices();
  busses = usb_get_busses();

  // Search OPTIMIZE's USB-GBA boot cable

  for (bus = busses; bus; bus = bus->next) {
    for (dev = bus->devices; dev; dev = dev->next) {
      dd = &dev->descriptor;
      if ((dd->idVendor == OPTUSB_VEND) && (dd->idProduct == OPTUSB_PROD))
        goto found;
     }
   }
  fprintf(stderr, "  Error: OPTIMIZE GBA-USB-bootcable is not found!\n");
  return 1;

  // Open the device

 found:
  optusb = usb_open(dev);
  if (optusb == NULL) {
    fprintf(stderr, "        Status = USB open error!\n");
    return 1;
   }

  usb_set_configuration(optusb, dev->config->bConfigurationValue);

  // Negotiate with the cable

  ret = read_magic(1);
  if (ret == 0x12345678 || ret == -1) {
    fprintf(stderr, "        Status = 1st negotiation error! (0x%08x)\n", ret);
    goto errexit;
   }

  send_magic(2, OPTUSB_MAGIC);
  for(;;) {
    ret = read_magic(2);
    if (ret == -1) {
      fprintf(stderr, "        Status = 2nd negotiation error! (0x%08x)\n", ret);
      goto errexit;
     }
    if (ret == OPTUSB_MAGIC) {
      break;
    }
  }

  send_magic(2, 0);
  for(;;) {
    ret = read_magic(0);
    if (ret == -1) {
      fprintf(stderr, "        Status = 3rd negotiation error! (0x%08x)\n", ret);
      goto errexit;
     }
    Sleep(100);
    if ((ret & (OPTUSB_READY|OPTUSB_ON)) == (OPTUSB_READY|OPTUSB_ON)){
      break;
    }
  }

  // Send the program file

  ret = read_magic(1);
  if (ret == 0x12345678 || ret == -1) {
    fprintf(stderr, "        Status = 4th negotiation error! (0x%08x)\n", ret);
    goto errexit;
   }
  ret = send_file((char*)buf, len);
  if (ret) {
    fprintf(stderr, "        Status = Transmission error! (0x%08x)\n", ret);
    goto errexit;
   }

  fprintf(stdout, "        Status = Successfully transferred.\n");

  usb_close(optusb);

  return 0;

 errexit:
  if (optusb != NULL)
    usb_close(optusb);
  return 1;
 }
